from datetime import timedelta
import logging
from logging import StreamHandler
from typing import Any, Dict

import cherrypy
from flask import Flask
from requestlogger import ApacheFormatter, WSGILogger
from werkzeug.contrib.fixers import ProxyFix

from .auth import setup_login

__all__ = ["cleanup_app", "create_app", "serve_app"]
logger = logging.getLogger("chaosplatform")


def create_app(config: Dict[str, Any]) -> Flask:
    """
    Create the Flask application.
    """
    app = Flask(__name__)
    app.url_map.strict_slashes = False
    app.debug = config.get("debug", False)

    logger = logging.getLogger('flask.app')
    logger.propagate = False

    app.config["SECRET_KEY"] = config["secret_key"]
    app.secret_key = app.config["SECRET_KEY"]

    app.config["JWT_SECRET_KEY"] = config["jwt"]["secret_key"]
    app.config["JWT_ACCESS_TOKEN_EXPIRES"] = timedelta(
        minutes=config["jwt"]["access_token_expires"])

    # OAUTH2
    oauth2_config = config["oauth2"]
    oauth2_gh_cfg = oauth2_config["github"]
    app.config["GITHUB_OAUTH_CLIENT_ID"] = oauth2_gh_cfg["client_id"]
    app.config["GITHUB_OAUTH_CLIENT_SECRET"] = oauth2_gh_cfg["client_secret"]

    app.config["CACHE_TYPE"] = config["cache"].get("type", "simple")
    if app.config["CACHE_TYPE"] == "redis":
        app.config["CACHE_REDIS_HOST"] = config["cache"]["host"]
        app.config["CACHE_REDIS_PORT"] = config["cache"]["port"]

    setup_login(app, from_session=True)

    return app


def cleanup_app(settings: Dict[str, Any]):
    pass


def serve_app(app: Flask, mount_point: str = '/',
              log_handler: StreamHandler = None):
    log_handler = log_handler or logging.StreamHandler()
    app.wsgi_app = ProxyFix(app.wsgi_app)
    wsgiapp = WSGILogger(
        app.wsgi_app, [log_handler], ApacheFormatter(),
        propagate=False)
    cherrypy.tree.graft(wsgiapp, mount_point)



from datetime import datetime
import logging
from logging import StreamHandler
import os
from typing import Any, Dict, Tuple
from unittest.mock import patch, MagicMock
import uuid
from uuid import UUID

import pytest

from chaosplatform.settings import load_settings

cur_dir = os.path.abspath(os.path.dirname(__file__))
fixtures_dir = os.path.join(cur_dir, "fixtures")
config_path = os.path.join(fixtures_dir, 'testconfig.toml')


@pytest.fixture
def config() -> Dict[str, Any]:
    return load_settings(config_path)



from typing import Any, Dict

try:
    from chaosplt_relational_storage import initialize_storage
    from chaosplt_relational_storage import configure_storage
    from chaosplt_relational_storage import release_storage
    HAS_RELATIONAL_STORAGE_PACKAGE = True
except ImportError:
    HAS_RELATIONAL_STORAGE_PACKAGE = False

__all__ = ["get_storage_driver", "configure_storage_driver",
           "release_storage_driver"]


def get_storage_driver(provider: str, config: Dict[str, Any]) -> object:
    if provider == "native":
        if not HAS_RELATIONAL_STORAGE_PACKAGE:
            raise RuntimeError(
                "'chaosplatform-relational-storage' not installed")
        return initialize_storage(config)


def configure_storage_driver(provider: str, storage_driver: object):
    if provider == "native":
        if not HAS_RELATIONAL_STORAGE_PACKAGE:
            raise RuntimeError(
                "'chaosplatform-relational-storage' not installed")
        configure_storage(storage_driver)


def release_storage_driver(provider: str, storage_driver: object):
    if provider == "native":
        if not HAS_RELATIONAL_STORAGE_PACKAGE:
            raise RuntimeError(
                "'chaosplatform-relational-storage' not installed")
        return release_storage(storage_driver)



import logging
from typing import Any, Dict, Tuple

from chaosplt_account.server import initialize_all as init_account, \
    release_all as release_account
from chaosplt_auth.server import initialize_all as init_auth, \
    release_all as release_auth
from chaosplt_experiment.server import initialize_all as init_experiment, \
    release_all as release_experiment
from chaosplt_scheduler.server import initialize_all as init_scheduler, \
    release_all as release_scheduler
from chaosplt_scheduling.server import initialize_all as init_scheduling, \
    release_all as release_scheduling
from chaosplt_grpc import create_grpc_server, start_grpc_server, \
    stop_grpc_server
import cherrypy
from flask import Flask
from grpc import Server as GRPCServer

from .app import create_app, cleanup_app, serve_app
from .api import create_api, cleanup_api, serve_api
from .cache import setup_cache
from .service import Services

__all__ = ["initialize_all", "release_all", "run_forever"]


def initialize_all(config: Dict[str, Any]) \
                   -> Tuple[Flask, Flask, Services, GRPCServer, Tuple, Tuple,
                            Tuple, Tuple, Tuple]:
    services = Services()

    access_log_handler = logging.StreamHandler()

    web_app = create_app(config)
    web_cache = setup_cache(web_app)

    api_app = create_api(config)
    api_cache = setup_cache(api_app)

    grpc_server = create_grpc_server(config["grpc"]["address"])
    start_grpc_server(grpc_server)

    account_resources = init_account(
        config["account"], web_app, api_app, services, grpc_server, web_cache,
        api_cache,
        web_mount_point="/account", api_mount_point="/account",
        access_log_handler=access_log_handler)
    auth_resources = init_auth(
        config["auth"], web_app, api_app, services, grpc_server, web_cache,
        api_cache,
        web_mount_point="/auth", api_mount_point="/auth",
        access_log_handler=access_log_handler)
    scheduler_resources = init_scheduler(
        config["scheduler"], services, grpc_server)
    scheduling_resources = init_scheduling(
        config["scheduling"], web_app, api_app, services, grpc_server,
        web_cache, api_cache,
        web_mount_point="/scheduling", api_mount_point="/scheduling",
        access_log_handler=access_log_handler)
    experiment_resources = init_experiment(
        config["experiment"], web_app, api_app, services, grpc_server,
        web_cache, api_cache,
        experiment_web_mount_point="/experi ment",
        experiment_api_mount_point="/experiments",
        execution_web_mount_point="/execution",
        execution_api_mount_point="/executions",
        access_log_handler=access_log_handler)

    serve_app(web_app, "/", access_log_handler)
    serve_api(api_app, "/api/v1", access_log_handler)

    return (
        web_app, api_app, services, grpc_server, auth_resources,
        account_resources, scheduler_resources, scheduling_resources,
        experiment_resources)


def release_all(services: Services, web_app: Flask, api_app: Flask,
                grpc_server: GRPCServer,
                auth_resources: Tuple, account_storage: Tuple,
                scheduler_resources: Tuple, scheduling_resources: Tuple,
                experiment_resources: Tuple):
    stop_grpc_server(grpc_server)
    cleanup_app(web_app)
    cleanup_api(api_app)
    release_scheduler(*scheduler_resources)
    release_account(*account_storage)
    release_auth(*auth_resources)
    release_scheduling(*scheduling_resources)
    release_experiment(*experiment_resources)


def run_forever(config: Dict[str, Any]):
    """
    Run and block until a signal is sent to the process.

    The application, services or gRPC server are all created and initialized
    when the application starts.
    """
    def run_stuff(config: Dict[str, Any]):
        resources = initialize_all(config)
        cherrypy.engine.subscribe(
            'stop', lambda: release_all(*resources),
            priority=20)

    cherrypy.engine.subscribe(
        'start', lambda: run_stuff(config), priority=80)

    cherrypy.engine.signals.subscribe()
    cherrypy.engine.start()
    cherrypy.engine.block()




# -*- coding: utf-8 -*-
from flask import Blueprint,  Flask
from flask_login import login_required

__all__ = ["register_views"]

plane = Blueprint("control_plane", __name__)


def register_views(app: Flask, prefix: str = '/'):
    prefix = prefix.rstrip("/")
    app.register_blueprint(plane, url_prefix="{}/".format(prefix))


@plane.route('')
@login_required
def index():
    return "hello"



import attr

__all__ = ["Services"]


@attr.s
class Services:
    auth: object = attr.ib(default=None)
    account: object = attr.ib(default=None)
    experiment: object = attr.ib(default=None)
    execution: object = attr.ib(default=None)
    scheduling: object = attr.ib(default=None)
    scheduler: object = attr.ib(default=None)
    worker: object = attr.ib(default=None)



# -*- coding: utf-8 -*-
from typing import Any, Dict

import cherrypy
import toml

__all__ = ["load_settings"]


def load_settings(toml_path: str) -> Dict[str, Any]:
    """
    """
    config = toml.load(toml_path)
    debug = config["chaosplatform"]["debug"]
    server_addr = config["chaosplatform"]["http"]["address"]
    host, port = server_addr.rsplit(":", 1)

    cherrypy.engine.unsubscribe('graceful', cherrypy.log.reopen_files)
    cherrypy.config.update({
        'server.socket_host': host,
        'server.socket_port': int(port),
        'engine.autoreload.on': False,
        'log.screen': debug,
        'log.access_file': '',
        'log.error_file': '',
        'environment': config["chaosplatform"].get("environment", "")
    })

    if "proxy" in config["chaosplatform"]["http"]:
        cherrypy.config.update({
            'tools.proxy.on': True,
            'tools.proxy.base': config["chaosplatform"]["http"]["proxy"]
        })

    return config["chaosplatform"]



from datetime import timedelta
import logging
from logging import StreamHandler
import os
from typing import Any, Dict

import cherrypy
from flask import Flask
from requestlogger import ApacheFormatter, WSGILogger
from werkzeug.contrib.fixers import ProxyFix

from .auth import setup_jwt, setup_login

__all__ = ["cleanup_api", "create_api", "serve_api"]
logger = logging.getLogger("chaosplatform")


def create_api(config: Dict[str, Any]) -> Flask:
    """
    Create the Flask application and initialize its resources.
    """
    app = Flask(__name__)

    app.url_map.strict_slashes = False
    app.debug = config.get("debug", False)

    logger = logging.getLogger('flask.app')
    logger.propagate = False

    app.config["SECRET_KEY"] = config["secret_key"]
    app.secret_key = app.config["SECRET_KEY"]

    app.config["JWT_SECRET_KEY"] = config["jwt"]["secret_key"]
    app.config["JWT_ACCESS_TOKEN_EXPIRES"] = timedelta(
        minutes=config["jwt"]["access_token_expires"])

    # OAUTH2
    app.config["GITHUB_OAUTH_CLIENT_ID"] = os.getenv("AUTH_GITHUB_CLIENT_ID")
    app.config["GITHUB_OAUTH_CLIENT_SECRET"] = os.getenv(
        "AUTH_GITHUB_CLIENT_SECRET")

    # OAUTH2
    oauth2_config = config["oauth2"]
    oauth2_gh_cfg = oauth2_config["github"]
    app.config["GITHUB_OAUTH_CLIENT_ID"] = oauth2_gh_cfg["client_id"]
    app.config["GITHUB_OAUTH_CLIENT_SECRET"] = oauth2_gh_cfg["client_secret"]

    app.config["CACHE_TYPE"] = config["cache"].get("type", "simple")
    if app.config["CACHE_TYPE"] == "redis":
        app.config["CACHE_REDIS_HOST"] = config["cache"]["host"]
        app.config["CACHE_REDIS_PORT"] = config["cache"]["port"]

    setup_jwt(app)
    setup_login(app, from_session=True, from_jwt=True)

    return app


def cleanup_api(settings: Dict[str, Any]):
    pass


def serve_api(app: Flask, mount_point: str = '/',
              log_handler: StreamHandler = None):
    log_handler = log_handler or logging.StreamHandler()
    app.wsgi_app = ProxyFix(app.wsgi_app)
    wsgiapp = WSGILogger(
        app.wsgi_app, [log_handler], ApacheFormatter(),
        propagate=False)
    cherrypy.tree.graft(wsgiapp, mount_point)
